from typing import Dict

from pydantic import BaseModel, Field

from otobo.domain_models.basic_auth_model import BasicAuth
from otobo.domain_models.ticket_operation import TicketOperation


class OTOBOClientConfig(BaseModel):
    base_url: str
    webservice_name: str
    auth: BasicAuth
    operation_url_map: dict[TicketOperation, str]





# tests/test_mappers.py
import pytest
from datetime import datetime

from mappers import build_ticket_create_request, parse_ticket_detail_output, build_ticket_update_request, \
    build_ticket_search_request, build_ticket_get_request
from otobo.domain_models.ticket_models import TicketBase, IdName, TicketSearch, Article
from otobo.models.ticket_models import OTOBOTicketBase, ArticleDetail, DynamicFieldItem, TicketDetailOutput
from otobo.models.request_models import TicketCreateRequest, TicketUpdateRequest, TicketGetRequest, TicketSearchRequest


def test_build_ticket_create_request_roundtrip():
    t = OTOBOTicketBase(
        title="Demo",
        queue=IdName(id=2, name="Raw"),
        state=IdName(name="new"),
        priority=IdName(name="3 normal"),
        type=IdName(name="Unclassified"),
        customer_user="user1",
        articles=[Article(subject="S", body="B", content_type="text/plain", from_addr="a@b.c", to_addr="x@y.z")],
    )
    req = build_ticket_create_request(t)
    assert isinstance(req, TicketCreateRequest)
    assert req.Ticket.Title == "Demo"
    assert req.Ticket.QueueID == 2
    assert req.Ticket.Queue == "Raw"
    assert req.Ticket.State == "new"
    assert req.Ticket.Priority == "3 normal"
    assert req.Ticket.Type == "Unclassified"
    articles = req.Article if isinstance(req.Article, list) else [req.Article] if req.Article else []
    assert articles[0].Subject == "S"
    assert articles[0].Body == "B"
    wire = TicketDetailOutput(
        Title=req.Ticket.Title,
        QueueID=req.Ticket.QueueID,
        Queue=req.Ticket.Queue,
        State=req.Ticket.State,
        Priority=req.Ticket.Priority,
        Type=req.Ticket.Type,
        CustomerUser=req.Ticket.CustomerUser,
        TicketID=111,
        TicketNumber="2025",
        Created=datetime.now().isoformat(),
        Changed=datetime.now().isoformat(),
        Article=req.Article,
        DynamicField=[],
    )
    back = parse_ticket_detail_output(wire)
    assert back.title == "Demo"
    assert back.queue == IdName(id=2, name="Raw")
    assert back.state == IdName(name="new")
    assert back.priority == IdName(name="3 normal")
    assert back.type == IdName(name="Unclassified")
    assert back.customer_user == "user1"
    assert len(back.articles) == 1
    assert back.articles[0].subject == "S"
    assert back.number == "2025"
    assert back.id == 111


def test_build_ticket_update_request_includes_ids_and_names():
    t = OTOBOTicketBase(
        id=123,
        number="TN-1",
        title="Updated",
        queue=IdName(id=5, name="Support"),
        state=IdName(id=1, name="open"),
        priority=IdName(id=3, name="3 normal"),
        type=IdName(id=2, name="Incident"),
        customer_user="user2",
    )
    req = build_ticket_update_request(t)
    assert isinstance(req, TicketUpdateRequest)
    assert req.TicketID == 123
    assert req.Ticket.Title == "Updated"
    assert req.Ticket.QueueID == 5
    assert req.Ticket.StateID == 1


def test_build_ticket_search_request_idname_lists():
    s = TicketSearch(
        numbers=["1001", "1002"],
        titles=["Demo"],
        queues=[IdName(id=2, name="Raw"), IdName(name="Support")],
        states=[IdName(name="open")],
        priorities=[IdName(id=4)],
        types=[IdName(name="Incident")],
        customer_users=["user1", "user2"],
        use_subqueues=True,
        limit=25,
    )
    req = build_ticket_search_request(s)
    assert isinstance(req, TicketSearchRequest)
    assert req.TicketNumber == ["1001", "1002"]
    assert req.Title == ["Demo"]
    assert sorted(req.QueueIDs or []) == [2]
    assert sorted(req.Queues or []) == ["Raw", "Support"]
    assert req.States == ["open"]
    assert req.PriorityIDs == [4]
    assert req.Types == ["Incident"]
    assert req.UseSubQueues == 1


def test_build_ticket_get_request_by_id_and_number():
    r1 = build_ticket_get_request(ticket_id=7)
    assert isinstance(r1, TicketGetRequest)
    assert r1.TicketID == 7


def test_parse_ticket_detail_output_handles_single_and_list_article():
    art = ArticleDetail(Subject="S1", Body="B1", ContentType="text/plain")
    wire_single = TicketDetailOutput(
        Title="A",
        TicketID=1,
        TicketNumber="N1",
        Article=art,
        DynamicField=[DynamicFieldItem(Name="K", Value="V")],
    )
    d1 = parse_ticket_detail_output(wire_single)
    assert len(d1.articles) == 1
    assert d1.articles[0].subject == "S1"
    wire_list = TicketDetailOutput(
        Title="A",
        TicketID=2,
        TicketNumber="N2",
        Article=[art, ArticleDetail(Subject="S2", Body="B2", ContentType="text/plain")],
        DynamicField=[],
    )
    d2 = parse_ticket_detail_output(wire_list)
    assert len(d2.articles) == 2
    assert [a.subject for a in d2.articles] == ["S1", "S2"]



import asyncio
import json
import logging
import random
import uuid
from http import HTTPMethod
from typing import Any, Optional

from httpx import AsyncClient
from pydantic import BaseModel, ValidationError

from otobo.domain_models.ticket_models import TicketBase, TicketSearch
from otobo.mappers import build_ticket_create_request, parse_ticket_detail_output, build_ticket_get_request, \
    build_ticket_update_request, build_ticket_search_request
from otobo.domain_models.otobo_client_config import OTOBOClientConfig
from otobo.domain_models.ticket_operation import TicketOperation
from otobo.models.request_models import (
    TicketSearchRequest,
    TicketUpdateRequest,
    TicketGetRequest, TicketCreateRequest,
)
from otobo.models.response_models import (
    TicketSearchResponse,
    TicketGetResponse,
    TicketResponse,
)
from otobo.models.ticket_models import TicketDetailOutput
from otobo.util.otobo_errors import OTOBOError


class OTOBOClient:
    def __init__(self, config: OTOBOClientConfig, client: AsyncClient | None = None, max_retries: int = 2):
        self.config = config
        self._client = client or AsyncClient()
        self.base_url = config.base_url.rstrip("/")
        self.webservice_name = config.webservice_name
        self.auth = config.auth
        self.operation_map = config.operation_url_map
        self.max_retries = max_retries
        self._logger = logging.getLogger(__name__)

    def _build_url(self, endpoint_name: str) -> str:
        return f"{self.base_url}/Webservice/{self.webservice_name}/{endpoint_name}"

    def _extract_error(self, payload: Any) -> Optional[OTOBOError]:
        if isinstance(payload, dict) and "Error" in payload:
            err = payload.get("Error") or {}
            return OTOBOError(str(err.get("ErrorCode", "")), str(err.get("ErrorMessage", "")))
        return None

    async def _send[T: BaseModel](
            self,
            method: HTTPMethod,
            operation: TicketOperation,
            response_model: type[T],
            data: dict[str, Any] | None = None,
    ) -> T:
        endpoint_name = self.operation_map[operation]
        url = self._build_url(endpoint_name)
        request_id = uuid.uuid4().hex
        payload = (self.auth.model_dump(by_alias=True, exclude_none=True) if isinstance(self.auth, BaseModel) else dict(
            self.auth)) | (data or {})

        self._logger.debug(f"[{request_id}] {method.value} {url} payload_keys={list(payload.keys())}")
        resp = await self._client.request(
            str(method.value),
            url,
            json=payload,
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )
        text = resp.text
        self._logger.debug(f"[{request_id}] status={resp.status_code} length={len(text)}")

        try:
            body = resp.json()
        except json.JSONDecodeError:
            self._logger.error(f"[{request_id}] invalid JSON response: {text[:500]}")
            resp.raise_for_status()
            raise

        api_err = self._extract_error(body)
        if api_err:
            self._logger.error(f"[{request_id}] OTOBO error {api_err.code}: {api_err.message}")
            raise api_err

        resp.raise_for_status()
        try:
            return response_model.model_validate(body, strict=False)
        except ValidationError as e:
            self._logger.error(f"[{request_id}] response validation error: {e}")
            return response_model.model_construct(**body)

    async def create_ticket(self, ticket: TicketBase) -> TicketBase:
        request: TicketCreateRequest = build_ticket_create_request(ticket)
        response: TicketResponse = await self._send(
            HTTPMethod.POST,
            TicketOperation.CREATE,
            TicketResponse,
            data=request.model_dump(exclude_none=True, by_alias=True),
        )
        if response.Ticket is None:
            raise RuntimeError("create returned no Ticket")
        return parse_ticket_detail_output(response.Ticket)

    async def get_ticket(self, ticket_id: int | str) -> TicketBase:
        request = build_ticket_get_request(int(ticket_id))
        response: TicketGetResponse = await self._send(
            HTTPMethod.POST,
            TicketOperation.GET,
            TicketGetResponse,
            data=request.model_dump(exclude_none=True, by_alias=True),
        )
        tickets = response.Ticket or []
        if len(tickets) != 1:
            raise RuntimeError(f"expected exactly one ticket, got {len(tickets)}")
        return parse_ticket_detail_output(
            tickets[0]
        )

    async def update_ticket(self, ticket: TicketBase) -> TicketBase:
        request = build_ticket_update_request(ticket)
        response: TicketResponse = await self._send(
            HTTPMethod.PUT,
            TicketOperation.UPDATE,
            TicketResponse,
            data=request.model_dump(exclude_none=True, by_alias=True),
        )
        if response.Ticket is None:
            raise RuntimeError("update returned no Ticket")
        return parse_ticket_detail_output(response.Ticket)

    async def search_tickets(self, ticket_search: TicketSearch) -> list[int]:
        request = build_ticket_search_request(ticket_search)
        response: TicketSearchResponse = await self._send(
            HTTPMethod.POST,
            TicketOperation.SEARCH,
            TicketSearchResponse,
            data=request.model_dump(exclude_none=True, by_alias=True),
        )
        return response.TicketID or []

    async def search_and_get(self, ticket_search: TicketSearch) -> list[
        TicketBase]:
        ids = await self.search_tickets(ticket_search)
        tasks = [self.get_ticket(i) for i in ids]
        return await asyncio.gather(*tasks)

    async def aclose(self) -> None:
        await self._client.aclose()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.aclose()



from __future__ import annotations

import copy
import re
from pathlib import Path
from typing import Any, Annotated, Dict, List, Literal

import typer
import yaml
from pydantic import BaseModel

from otobo import TicketOperation


class NoAliasDumper(yaml.SafeDumper):
    def ignore_aliases(self, data):
        return True


app = typer.Typer(
    add_completion=False,
    help="Generate secure OTOBO/Znuny web service YAML.",
    context_settings={"help_option_names": ["-h", "--help"]},
)


class RouteMappingConfig(BaseModel):
    Route: str
    RequestMethod: List[str]
    ParserBackend: Literal["JSON"] = "JSON"


class ProviderOperationConfig(BaseModel):
    Type: str
    Description: str
    IncludeTicketData: Literal["0", "1"]
    MappingInbound: Dict[str, Any]
    MappingOutbound: Dict[str, Any]


class OperationSpec(BaseModel):
    op: TicketOperation
    route: str
    description: str
    methods: List[str]
    include_ticket_data: Literal["0", "1"]


class WebServiceGenerator:
    DEFAULT_SPECS: dict[TicketOperation, OperationSpec] = {
        TicketOperation.CREATE: OperationSpec(
            op=TicketOperation.CREATE,
            route="ticket-create",
            description="Creates a new ticket.",
            methods=["POST"],
            include_ticket_data="1",
        ),
        TicketOperation.GET: OperationSpec(
            op=TicketOperation.GET,
            route="ticket-get",
            description="Retrieves a ticket by its ID.",
            methods=["POST"],
            include_ticket_data="1",
        ),
        TicketOperation.SEARCH: OperationSpec(
            op=TicketOperation.SEARCH,
            route="ticket/search",
            description="Searches for tickets based on criteria.",
            methods=["POST"],
            include_ticket_data="0",
        ),
        TicketOperation.UPDATE: OperationSpec(
            op=TicketOperation.UPDATE,
            route="ticket/update",
            description="Updates an existing ticket.",
            methods=["PUT"],
            include_ticket_data="1",
        ),
    }

    def _create_operation_configs(self, s: OperationSpec, inbound: Dict[str, Any]) -> dict[str, dict]:
        return ProviderOperationConfig(
            Type=s.op.type,
            Description=s.description,
            IncludeTicketData=s.include_ticket_data,
            MappingInbound=inbound,
            MappingOutbound=self._outbound_mapping(),
        ).model_dump()

    def generate_yaml(
            self,
            webservice_name: str,
            enabled_operations: List[TicketOperation],
            restricted_user: str | None = None,
            framework_version: str = "11.0.0",
    ) -> str:
        name = self._validate_name(webservice_name)
        specs = [self.DEFAULT_SPECS[o] for o in enabled_operations if o in self.DEFAULT_SPECS]
        if not specs:
            raise ValueError("No operations enabled.")
        inbound_base = self._inbound_mapping(restricted_user)
        description = self._description(name, restricted_user)
        route_mapping: dict[str, dict] = {}
        operations: dict[str, dict] = {}
        for s in specs:
            inbound = copy.deepcopy(inbound_base)
            route_mapping[s.provider_name] = RouteMappingConfig(
                Route=f"/{s.route}",
                RequestMethod=s.methods,
            ).model_dump()
            operations[s.provider_name] = self._create_operation_configs(s, inbound)
        data = {
            "Debugger": {"DebugThreshold": "debug", "TestMode": "0"},
            "Description": description,
            "FrameworkVersion": framework_version,
            "Provider": {
                "Transport": {
                    "Type": "HTTP::REST",
                    "Config": {
                        "MaxLength": "1000000",
                        "KeepAlive": "",
                        "AdditionalHeaders": "",
                        "RouteOperationMapping": route_mapping,
                    },
                },
                "Operation": operations,
            },
            "RemoteSystem": "",
            "Requester": {"Transport": {"Type": ""}},
        }
        return yaml.dump(data, sort_keys=False, indent=2, Dumper=NoAliasDumper, explicit_start=True)

    def _validate_name(self, name: str) -> str:
        if not name:
            raise ValueError("Webservice name cannot be empty.")
        if not re.fullmatch(r"[A-Za-z][A-Za-z0-9_-]*", name):
            raise ValueError("Name must start with a letter and contain only A–Z, a–z, 0–9, _ or -.")
        return name

    def _description(self, name: str, user: str | None) -> str:
        if user:
            return f"Webservice for '{name}'. Restricted to user '{user}'."
        return f"Webservice for '{name}'. Accessible by all permitted agents."

    def _inbound_mapping(self, restricted_user: str | None) -> Dict[str, Any]:
        if restricted_user:
            return {
                "Type": "Simple",
                "Config": {
                    "KeyMapDefault": {"MapType": "Keep", "MapTo": ""},
                    "KeyMapExact": {"UserLogin": "UserLogin"},
                    "ValueMap": {"UserLogin": {"ValueMapRegEx": {".*": restricted_user}}},
                    "ValueMapDefault": {"MapType": "Keep", "MapTo": ""},
                },
            }
        return {
            "Type": "Simple",
            "Config": {
                "KeyMapDefault": {"MapType": "Keep", "MapTo": ""},
                "ValueMapDefault": {"MapType": "Keep", "MapTo": ""},
            },
        }

    def _outbound_mapping(self) -> Dict[str, Any]:
        return {
            "Type": "Simple",
            "Config": {
                "KeyMapDefault": {"MapTo": "", "MapType": "Keep"},
                "ValueMapDefault": {"MapTo": "", "MapType": "Keep"},
            },
        }


@app.command()
def generate(
        name: Annotated[str, typer.Option("--name", rich_help_panel="Required")],
        enable_ticket_get: Annotated[bool, typer.Option("--enable-ticket-get", rich_help_panel="Operations")] = False,
        enable_ticket_search: Annotated[
            bool, typer.Option("--enable-ticket-search", rich_help_panel="Operations")] = False,
        enable_ticket_create: Annotated[
            bool, typer.Option("--enable-ticket-create", rich_help_panel="Operations")] = False,
        enable_ticket_update: Annotated[
            bool, typer.Option("--enable-ticket-update", rich_help_panel="Operations")] = False,
        allow_user: Annotated[
            str | None, typer.Option("--allow-user", metavar="USERNAME", rich_help_panel="Auth")] = None,
        allow_all_agents: Annotated[bool, typer.Option("--allow-all-agents", rich_help_panel="Auth")] = False,
        version: Annotated[str, typer.Option("--version", rich_help_panel="Optional")] = "11.0.0",
        file: Annotated[str | None, typer.Option("--file", metavar="FILENAME", rich_help_panel="Optional")] = None,
):
    if not (allow_user or allow_all_agents):
        typer.secho("Error: You must specify an authentication mode.", fg=typer.colors.RED)
        raise typer.Exit(code=1)
    if allow_user and allow_all_agents:
        typer.secho("Error: --allow-user and --allow-all-agents are mutually exclusive.", fg=typer.colors.RED)
        raise typer.Exit(code=1)
    enabled: list[TicketOperation] = [
        op
        for op, ok in [
            (TicketOperation.GET, enable_ticket_get),
            (TicketOperation.SEARCH, enable_ticket_search),
            (TicketOperation.CREATE, enable_ticket_create),
            (TicketOperation.UPDATE, enable_ticket_update),
        ]
        if ok
    ]
    gen = WebServiceGenerator()
    try:
        out = gen.generate_yaml(
            webservice_name=name,
            enabled_operations=enabled,
            restricted_user=allow_user if allow_user else None,
            framework_version=version,
        )
        if file:
            Path(file).write_text(out, encoding="utf-8")
            typer.secho("Successfully generated webservice configuration.", fg=typer.colors.GREEN)
            typer.secho(f"File: {file}")
        else:
            typer.secho("--- Generated YAML ---", bold=True)
            print(out)
    except ValueError as e:
        typer.secho(f"Error: {e}", fg=typer.colors.RED)
        raise typer.Exit(code=1)


if __name__ == "__main__":
    app()



from abc import abstractmethod, ABC
from datetime import datetime

from pydantic import BaseModel


class IdName(BaseModel):
    id: int | None = None
    name: str | None = None


class Article(BaseModel):
    from_addr: str | None = None
    to_addr: str | None = None
    subject: str | None = None
    body: str | None = None
    content_type: str | None = None
    created_at: datetime | None = None
    changed_at: datetime | None = None
    article_id: int | None = None
    article_number: int | None = None


class TicketBase(BaseModel, ABC):
    number: str | None = None
    title: str | None = None
    queue: IdName | None = None
    state: IdName | None = None
    priority: IdName | None = None
    type: IdName | None = None
    owner: IdName | None = None
    customer_id: str | None = None
    customer_user: str | None = None
    created_at: datetime | None = None
    changed_at: datetime | None = None

    @abstractmethod
    def get_articles(self):
        pass


class TicketCreate(TicketBase):
    article: Article | None = None

    def get_articles(self):
        return [self.article] if self.article else []


class TicketUpdate(TicketBase):
    id: int | None = None
    article: Article | None = None

    def get_articles(self):
        return [self.article] if self.article else []


class Ticket(TicketBase):
    id: int | None = None
    articles: list[Article] | None = None

    def get_articles(self):
        return self.articles or []


class TicketSearch(BaseModel):
    numbers: list[str] | None = None
    titles: list[str] | None = None
    queues: list[IdName] | None = None
    states: list[IdName] | None = None
    priorities: list[IdName] | None = None
    types: list[IdName] | None = None
    customer_users: list[str] | None = None
    use_subqueues: bool = False
    limit: int = 50



from typing import Union, List, Optional

from pydantic import BaseModel

from otobo.models.ticket_models import TicketDetailOutput
from otobo.util.otobo_errors import OTOBOError

class TicketResponse(BaseModel):
    Ticket: Optional[TicketDetailOutput] = None


class TicketGetResponse(BaseModel):
    Ticket: list[TicketDetailOutput]


class TicketSearchResponse(BaseModel):
    TicketID: List[int] | None = None



from typing import Optional, Union, List, Dict, Literal

from pydantic import BaseModel, Field, field_validator

from models.ticket_models import OTOBOTicketBase
from otobo.models.ticket_models import ArticleDetail, DynamicFieldItem


class AuthData(BaseModel):
    UserLogin: str = Field(..., description="Agent login for authentication")
    Password: str = Field(..., description="Agent password for authentication")


class TicketSearchRequest(BaseModel):
    TicketNumber: Optional[Union[str, List[str]]] = None
    Title: Optional[Union[str, List[str]]] = None
    Queues: Optional[List[str]] = None
    QueueIDs: Optional[List[int]] = None
    UseSubQueues: Optional[bool] = False
    Types: Optional[List[str]] = None
    TypeIDs: Optional[List[int]] = None
    States: Optional[List[str]] = None
    StateIDs: Optional[List[int]] = None
    Priorities: Optional[List[str]] = None
    PriorityIDs: Optional[List[int]] = None
    Limit: int = 0
    SearchLimit: int = 0


class TicketGetRequest(BaseModel):
    TicketID: Optional[int] = None
    DynamicFields: int = 1
    Extended: int = 1
    AllArticles: int = 1
    ArticleSenderType: Optional[List[str]] = None
    ArticleOrder: Literal["ASC", "DESC"] = 'ASC'
    ArticleLimit: int = 5
    Attachments: int = 0
    GetAttachmentContents: int = 1
    HTMLBodyAsAttachment: int = 1



class TicketCreateRequest(BaseModel):
    Ticket: OTOBOTicketBase | None = None
    Article: ArticleDetail | list[ArticleDetail] | None = None
    DynamicField: list[DynamicFieldItem] | None = None


class TicketUpdateRequest(TicketCreateRequest):
    TicketID: Optional[int] = None
    TicketNumber: Optional[str] = None



from typing import Any

from otobo.domain_models.ticket_models import TicketUpdate, Ticket, TicketCreate
from otobo.domain_models.ticket_models import Article, IdName, TicketBase, TicketSearch
from otobo.models.request_models import TicketCreateRequest, TicketUpdateRequest, TicketSearchRequest, TicketGetRequest
from otobo.models.ticket_models import DynamicFieldItem, ArticleDetail, TicketDetailOutput, OTOBOTicketBase


def _dynamic_fields_dict_to_items(dynamic_fields: dict[str, Any]) -> list[DynamicFieldItem]:
    return [DynamicFieldItem(Name=key, Value=value) for key, value in dynamic_fields.items()]


def _dynamic_fields_items_to_dict(dynamic_items: list[DynamicFieldItem] | None) -> dict[str, Any]:
    return {item.Name: item.Value for item in dynamic_items or []}


def _article_domain_to_wire(article: Article) -> ArticleDetail:
    return ArticleDetail(
        From=article.from_addr,
        To=article.to_addr,
        Subject=article.subject,
        Body=article.body,
        ContentType=article.content_type,
    )


def _article_wire_to_domain(article_wire: ArticleDetail) -> Article:
    return Article(
        from_addr=article_wire.From,
        to_addr=article_wire.To,
        subject=article_wire.Subject,
        body=article_wire.Body,
        content_type=article_wire.ContentType,
        created_at=article_wire.CreateTime,
        changed_at=article_wire.ChangeTime,
        article_id=article_wire.ArticleID,
        article_number=article_wire.ArticleNumber,
    )


def _to_idname(id_value: int | None, name_value: str | None) -> IdName | None:
    if id_value is None and name_value is None:
        return None
    return IdName(id=id_value, name=name_value)


def _split_idname_sequence(items: list[IdName] | None) -> tuple[list[int] | None, list[str] | None]:
    if not items:
        return None, None
    id_list = [x.id for x in items if x.id is not None] or None
    name_list = [x.name for x in items if x.name is not None] or None
    return id_list, name_list


def parse_ticket_detail_output(ticket_wire: TicketDetailOutput | dict) -> Ticket:
    if isinstance(ticket_wire, dict):
        if 'Article' in ticket_wire:
            if isinstance(ticket_wire['Article'], dict):
                ticket_wire['Article'] = [ticket_wire['Article']]
        ticket_wire = TicketDetailOutput.model_validate(ticket_wire, strict=False)
    if isinstance(ticket_wire.Article, dict):
        ticket_wire.Article = [ArticleDetail.model_validate(ticket_wire.Article)]
    if 'Article' in ticket_wire or ticket_wire.Article:
        wire_articles = ticket_wire.Article \
            if isinstance(ticket_wire.Article, list) \
            else [ticket_wire.Article] if ticket_wire.Article else []
    else:
        wire_articles = []
    return Ticket(
        id=ticket_wire.TicketID,
        number=ticket_wire.TicketNumber,
        title=ticket_wire.Title,
        queue=_to_idname(ticket_wire.QueueID, ticket_wire.Queue),
        state=_to_idname(ticket_wire.StateID, ticket_wire.State),
        priority=_to_idname(ticket_wire.PriorityID, ticket_wire.Priority),
        type=_to_idname(ticket_wire.TypeID, ticket_wire.Type),
        owner=_to_idname(ticket_wire.OwnerID, ticket_wire.Owner),
        customer_id=ticket_wire.CustomerID,
        customer_user=ticket_wire.CustomerUser,
        created_at=ticket_wire.Created,
        changed_at=ticket_wire.Changed,
        articles=[_article_wire_to_domain(a) for a in wire_articles],
    )


def build_ticket_base(ticket: TicketBase) -> OTOBOTicketBase | None:
    assert isinstance(ticket, TicketBase)
    def id_name(v: IdName | None) -> tuple[int | None, str | None]:
        return (v.id, v.name) if v else (None, None)

    def has_any_attribute_set(otobo_ticket_base: OTOBOTicketBase) -> bool:
        return bool(otobo_ticket_base.model_dump(exclude_none=True))

    queue_id, queue_name = id_name(ticket.queue)
    state_id, state_name = id_name(ticket.state)
    priority_id, priority_name = id_name(ticket.priority)
    type_id, type_name = id_name(ticket.type)

    ticket: OTOBOTicketBase = OTOBOTicketBase(
        Title=ticket.title,
        QueueID=queue_id,
        Queue=queue_name,
        StateID=state_id,
        State=state_name,
        PriorityID=priority_id,
        Priority=priority_name,
        CustomerUser=ticket.customer_user,
        TypeID=type_id,
        Type=type_name,
    )

    if has_any_attribute_set(ticket):
        return ticket
    return None


def build_ticket_create_request(ticket_domain: TicketCreate) -> TicketCreateRequest:
    ticket_base = build_ticket_base(ticket_domain)
    article_wire = _article_domain_to_wire(ticket_domain.article) if ticket_domain.article else None
    return TicketCreateRequest(Ticket=ticket_base, Article=article_wire)


def build_ticket_update_request(ticket_domain: TicketUpdate) -> TicketUpdateRequest:
    ticket_base = build_ticket_base(ticket_domain)
    article_wire = _article_domain_to_wire(ticket_domain.article) if ticket_domain.article else None
    return TicketUpdateRequest(
        Ticket=ticket_base,
        Article=article_wire,
        TicketID=ticket_domain.id,
        TicketNumber=ticket_domain.number,
    )


def build_ticket_search_request(search_model: TicketSearch) -> TicketSearchRequest:
    queue_ids, queue_names = _split_idname_sequence(search_model.queues)
    state_ids, state_names = _split_idname_sequence(search_model.states)
    priority_ids, priority_names = _split_idname_sequence(search_model.priorities)
    type_ids, type_names = _split_idname_sequence(search_model.types)
    return TicketSearchRequest(
        TicketNumber=search_model.numbers,
        Title=search_model.titles,
        Queues=queue_names,
        QueueIDs=queue_ids,
        States=state_names,
        StateIDs=state_ids,
        Priorities=priority_names,
        PriorityIDs=priority_ids,
        Types=type_names,
        TypeIDs=type_ids,
        UseSubQueues=search_model.use_subqueues,
        Limit=search_model.limit,
    )


def build_ticket_get_request(ticket_id: int) -> TicketGetRequest:
    return TicketGetRequest(TicketID=ticket_id)



from typing import Any, List, Union, Optional
from pydantic import BaseModel


class OTOBOTicketBase(BaseModel):
    Title: Optional[str] = None
    QueueID: Optional[int] = None
    Queue: Optional[str] = None
    StateID: Optional[int] = None
    State: Optional[str] = None
    PriorityID: Optional[int] = None
    Priority: Optional[str] = None
    OwnerID: Optional[int] = None
    Owner: Optional[str] = None
    CustomerUser: Optional[str] = None
    TicketID: Optional[int] = None
    TicketNumber: Optional[str] = None
    Type: Optional[str] = None
    TypeID: Optional[int] = None
    CustomerID: Optional[str] = None
    CustomerUserID: Optional[str] = None
    CreateBy: Optional[int] = None
    ChangeBy: Optional[int] = None
    Created: Optional[str] = None
    Changed: Optional[str] = None



class DynamicFieldItem(BaseModel):
    Name: str
    Value: Optional[Any] = None


class ArticleDetail(BaseModel):
    ArticleID: Optional[int] = None
    ArticleNumber: Optional[int] = None
    From: Optional[str] = None
    Subject: Optional[str] = None
    Body: Optional[str] = None
    ContentType: Optional[str] = None
    CreateTime: Optional[str] = None
    ChangeTime: Optional[str] = None
    To: Optional[str] = None
    MessageID: Optional[str] = None
    ChangeBy: Optional[int] = None
    CreateBy: Optional[int] = None


class TicketDetailOutput(OTOBOTicketBase):
    Article: List[ArticleDetail] | ArticleDetail | None = None
    DynamicField: List[DynamicFieldItem] | None = None





#!/usr/bin/env python3
from __future__ import annotations

import json
import re
import secrets
import string
from pathlib import Path
from typing import List, Optional, Dict, Any

import requests
import typer

from setup_webservices import WebServiceGenerator

app = typer.Typer(add_completion=False, context_settings={"help_option_names": ["-h", "--help"]})

OPERATIONS: Dict[str, str] = {
    "get": "TicketGet",
    "search": "TicketSearch",
    "create": "TicketCreate",
    "update": "TicketUpdate",
}


def _generate_password(length: int = 20) -> str:
    alphabet = string.ascii_letters + string.digits
    return "".join(secrets.choice(alphabet) for _ in range(length))


def _slug(s: str) -> str:
    s2 = re.sub(r"[^A-Za-z0-9]+", "-", s).strip("-")
    return s2 or "ws"


def _email_local(s: str) -> str:
    return re.sub(r"[^a-z0-9._+-]", "", s.lower())


def _write_file(path: Path, data: str, force: bool) -> None:
    if path.exists() and not force:
        typer.secho(f"File exists: {path}. Use --force to overwrite.", fg=typer.colors.RED)
        raise typer.Exit(code=1)
    path.write_text(data, encoding="utf-8")
    typer.secho(f"Wrote {path}", fg=typer.colors.GREEN)


def _prompt_yes(msg: str, default: bool = True) -> bool:
    return typer.confirm(msg, default=default)


def _http_post(url: str, payload: Dict[str, Any]) -> Dict[str, Any]:
    r = requests.post(url, json=payload, timeout=30)
    try:
        return r.json()
    except Exception:
        return {"status": r.status_code, "text": r.text}


@app.command()
def bootstrap(
        base_url: str = typer.Option(..., "--base-url", "-u",
                                     help="Server base URL, e.g. http://host/otobo/nph-genericinterface.pl"),
        name: str = typer.Option(..., "--name", "-n", help="Webservice name (letters, numbers, dashes)"),
        op: List[str] = typer.Option(..., "--op", "-o", help="Repeat for each: get, search, create, update",
                                     case_sensitive=False),
        restrict_user: Optional[str] = typer.Option(None, "--user", help="Agent login to restrict access to"),
        first_name: Optional[str] = typer.Option(None, "--first-name"),
        last_name: Optional[str] = typer.Option(None, "--last-name"),
        email: Optional[str] = typer.Option(None, "--email"),
        version: str = typer.Option("11.0.0", "--version", help="FrameworkVersion"),
        out: Optional[Path] = typer.Option(None, "--out", "-f", help="Output YAML file; defaults to <name>.yml"),
        write_env: bool = typer.Option(False, "--write-env/--no-write-env", help="Write .env with server and username"),
        env_file: Path = typer.Option(Path(".env"), "--env-file"),
        force: bool = typer.Option(False, "--force", help="Overwrite existing files"),
        run_test: bool = typer.Option(True, "--test", help="Attempt live HTTP test after setup"),
):
    ops_clean = [s.lower().strip() for s in op if s.lower().strip() in OPERATIONS]
    if not ops_clean:
        typer.secho("No valid operations provided.", fg=typer.colors.RED)
        raise typer.Exit(code=1)

    slug_name = _slug(name)
    default_first = name
    default_last = "User"
    default_email = f"{_email_local(slug_name)}@localhost.com"

    create_agent = False
    if not restrict_user:
        create_agent = _prompt_yes("Create a new Agent for this webservice?", True)
        if create_agent:
            restrict_user = typer.prompt("Agent login", default=f"{slug_name}_ws".lower())
            first_name = typer.prompt("First name", default=default_first)
            last_name = typer.prompt("Last name", default=default_last)
            email = typer.prompt("Email", default=default_email)
        else:
            restrict_user = typer.prompt("Existing agent login to use", default="root@localhost")
    else:
        if not first_name:
            first_name = default_first
        if not last_name:
            last_name = default_last
        if not email:
            email = default_email

    enabled_ops: dict[str, None] = {OPERATIONS[o]: None for o in ops_clean}
    generator = WebServiceGenerator()
    yaml_content = generator.generate_yaml(
        webservice_name=name,
        enabled_operations=enabled_ops,
        restricted_user=restrict_user,
        framework_version=version,
    )

    target = out or Path(f"{slug_name}.yml")
    _write_file(target, yaml_content, force)

    typer.echo("Command to add webservice:")
    typer.echo(f"otobo.Console.pl Admin::WebService::Add --Name {name} --Config {target}")

    created_password = None
    if create_agent:
        created_password = _generate_password()
        typer.echo("Command to create agent:")
        typer.echo(
            f"otobo.Console.pl Admin::User::Add --UserFirstname \"{first_name}\" --UserLastname \"{last_name}\" "
            f"--UserLogin \"{restrict_user}\" --UserPassword \"{created_password}\" --UserEmail \"{email}\""
        )

    make_gq = _prompt_yes("Create a new Group and Queue for testing?", True)
    group_name = None
    queue_name = None
    if make_gq:
        group_name = typer.prompt("Group name", default=f"{slug_name}-group")
        queue_name = typer.prompt("Queue name", default=f"{slug_name}::Test")
        typer.echo("Commands to create group and queue:")
        typer.echo(f"otobo.Console.pl Admin::Group::Add --Name \"{group_name}\" --Valid 1")
        typer.echo(f"otobo.Console.pl Admin::Queue::Add --Name \"{queue_name}\" --Group \"{group_name}\" --Valid 1")
    else:
        group_name = typer.prompt("Existing group name", default="users")
        queue_name = typer.prompt("Existing queue name", default="Raw")

    grant_rights = _prompt_yes("Grant agent rights on the group (ro,note,create,move_into)?", True)
    if grant_rights:
        typer.echo("Command to link agent to group with permissions:")
        typer.echo(
            f"otobo.Console.pl Admin::Group::UserLink --group-name \"{group_name}\" --user-name \"{restrict_user}\" --permission \"ro,move_into,create,note\""
        )

    if write_env:
        lines = [f"OTOBO_SERVER_URL={base_url}", f"OTOBO_WEBSERVICE_NAME={name}", f"OTOBO_QUEUE={queue_name}"]
        lines.append(f"OTOBO_USERNAME={restrict_user}")
        _write_file(env_file, "\n".join(lines) + "\n", force)

    if not run_test:
        return

    op_urls = {k: f"{base_url.rstrip('/')}/Webservice/{name}/{v}" for k, v in OP_PATHS.items() if k in enabled_ops}

    auth_login = restrict_user
    auth_password = created_password or typer.prompt("Password for the agent to authenticate HTTP calls",
                                                     hide_input=True)

    created_ticket_number: Optional[str] = None
    created_ticket_id: Optional[int] = None

    if "TicketCreate" in enabled_ops:
        payload_create = {
            "UserLogin": auth_login,
            "Password": auth_password,
            "Ticket": {
                "Title": f"{name} setup test",
                "Queue": queue_name,
                "State": "open",
                "Priority": "3 normal",
                "CustomerUser": "test@example.com",
            },
            "Article": {
                "Subject": "Setup test",
                "Body": "Initial ticket via webservice",
                "MimeType": "text/plain",
                "Charset": "utf-8",
            },
        }
        res_c = _http_post(op_urls.get("TicketCreate", ""), payload_create)
        typer.echo(f"Create response: {json.dumps(res_c, ensure_ascii=False)}")
        created_ticket_number = res_c.get("TicketNumber")
        created_ticket_id = res_c.get("TicketID")

    if "TicketUpdate" in enabled_ops:
        if not created_ticket_id:
            if created_ticket_number and "TicketSearch" in enabled_ops:
                payload_search = {"UserLogin": auth_login, "Password": auth_password,
                                  "TicketNumber": created_ticket_number}
                res_s = _http_post(op_urls.get("TicketSearch", ""), payload_search)
                ids = res_s.get("TicketIDs") or []
                if isinstance(ids, list) and ids:
                    created_ticket_id = ids[0]
            if not created_ticket_id:
                num = typer.prompt("TicketNumber to update (no create op available or failed)",
                                   default=created_ticket_number or "")
                payload_search2 = {"UserLogin": auth_login, "Password": auth_password, "TicketNumber": num}
                res_s2 = _http_post(op_urls.get("TicketSearch", ""), payload_search2)
                ids2 = res_s2.get("TicketIDs") or []
                if isinstance(ids2, list) and ids2:
                    created_ticket_id = ids2[0]
        if created_ticket_id:
            payload_update = {
                "UserLogin": auth_login,
                "Password": auth_password,
                "TicketID": created_ticket_id,
                "Ticket": {"Priority": "4 high"},
                "Article": {"Subject": "Update", "Body": "Update via webservice"},
            }
            res_u = _http_post(op_urls.get("TicketUpdate", ""), payload_update)
            typer.echo(f"Update response: {json.dumps(res_u, ensure_ascii=False)}")

    tn_for_get = created_ticket_number
    if not tn_for_get:
        tn_for_get = typer.prompt("TicketNumber to fetch", default=created_ticket_number or "")
    if tn_for_get and "TicketSearch" in enabled_ops:
        res_s3 = _http_post(op_urls.get("TicketSearch", ""),
                            {"UserLogin": auth_login, "Password": auth_password, "TicketNumber": tn_for_get})
        typer.echo(f"Search response: {json.dumps(res_s3, ensure_ascii=False)}")
        ids3 = res_s3.get("TicketIDs") or []
        tid = ids3[0] if isinstance(ids3, list) and ids3 else None
        if tid and "TicketGet" in enabled_ops:
            res_g = _http_post(op_urls.get("TicketGet", ""),
                               {"UserLogin": auth_login, "Password": auth_password, "TicketID": tid})
            typer.echo(f"Get response: {json.dumps(res_g, ensure_ascii=False)}")


if __name__ == "__main__":
    app()



from enum import Enum
from typing import Any


class TicketOperation(Enum):
    CREATE = ("TicketCreate", "Ticket::TicketCreate")
    SEARCH = ("TicketSearch", "Ticket::TicketSearch")
    GET = ("TicketGet", "Ticket::TicketGet")
    UPDATE = ("TicketUpdate", "Ticket::TicketUpdate")

    def __new__(cls, name: str, operation_type: str):
        obj = object.__new__(cls)
        obj._value_ = name
        obj.operation_type = operation_type
        return obj

    @property
    def type(self) -> str:
        return self.operation_type

    def __hash__(self) -> int:
        return hash(self.value)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, TicketOperation):
            return self.value == other.value
        if isinstance(other, str):
            return self.value == other
        return False
